/* 
 * NATOW - Tank to tank warfare in a 3D fully destructable world
 *
 * Copyright (C) 2000-2001
 *                     Robert Ancell  <bob27@users.sourceforge.net>
 *                     Yannick Perret <hexasoft@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* draw_map.c
   Contains the routines used to draw the landscape  */

#include <stdio.h>
#include <GL/gl.h>
#include <math.h>

#include "draw_map.h"
#include "world.h"
#include "game.h"
#include "update.h"
#include "map.h"
#include "satellite.h"
#include "text.h"
#include "tools.h"

static void draw_skirt(Game *game);
static void draw_ground(Game *game);
static void draw_lakes(Game *game);
static void draw_sky(Game *game);
static void draw_map_loop(Map *map, int, int, int, int, int, int, int, int, int, int);
static void draw_outer_map_loop(Map *map, int, int, int, int, int, int, int, int, int);
static void draw_map_segment(Map *map, int, int, int, int, int);
static void draw_xstrip(Map *map, int, int, int, int, int);
static void draw_xjoiner(Map *map, int, int, int, int, int, int, int, int);
static void draw_xjoiner_reverse(Map *map, int, int, int, int, int, int, int, int);
static void draw_yjoiner(Map *map, int, int, int, int, int, int, int, int);
static void draw_yjoiner_reverse(Map *map, int, int, int, int, int, int, int, int);
static void mapVertex(Map *map, int, int);
static void mapVertexs(Map *map, int, int, int);

/* init the drawing module */
double pc_cos[17]={0.};
double pc_sin[17]={0.};

void init_draw_map(void)
{
  float angle;
  int i;

  Nprintf("Initializing map-drawing module...\n");
  angle = 0.;
  for(i=0; i<17; i++)
    {
    pc_cos[i] = cos(angle);
    pc_sin[i] = sin(angle);
    angle += 2.*3.14159/16.;
    }
  Nprintf("Done.\n");
}

/* Draw the fixed features */
void draw_map(Game *game)
{
  /* Enable texturing if necessary */
  glPushAttrib(GL_TEXTURE_BIT);
  if(game->world->is_texture)
    glEnable(GL_TEXTURE_2D);
  glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

  /* Draw the sky (do it BEFORE the rest) */
  if(game->world->is_sky)
    draw_sky(game);

  /* draw sun, moons... */
  if(game->world->sun != NULL)
    draw_satellite(game->world->sun);
  if(game->world->moon1 != NULL)
    draw_satellite(game->world->moon1);
  if(game->world->moon2 != NULL)
    draw_satellite(game->world->moon2);

  /* Draw the ground */
  if(game->world->is_texture)
    glEnable(GL_TEXTURE_2D);
  glBindTexture(GL_TEXTURE_2D, game->map->texture);
  draw_ground(game);

  /* Extend the ground (todo: change this in GL List) */
  draw_skirt(game);

  /* Draw the trees */
  /* NOTE: to be added */
  /*if(world->is_trees)
    draw_trees(world);*/

  glPopAttrib();

  glDisable(GL_TEXTURE_2D);
}

/* Draw a 'skirt' around the map, so map appears to extend to horizon */
static void draw_skirt(Game *game)
{
  Map *map = game->map;
  float w, h, oldw, oldh;
  float theta, sin_theta, cos_theta, old_sin_theta, old_cos_theta;
  float width = map->width, height = map->height;
  float half_width = 0.5 * width, half_height = 0.5 * height;
  float radius;
  GLfloat bg[4]={0.0, 0.0, 0.0, 1.0};

  radius = 4.0 * nMAX(width, height);

  glDisable(GL_TEXTURE_2D);
  glDisable(GL_BLEND);
  glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, bg);
  glMaterialf(GL_FRONT, GL_SHININESS, 128.0f);
  glColor4fv(bg);
/*
  glNormal3f(0., 1., 0.);
  glBegin(GL_TRIANGLES);
  for(i=0; i<16; i++)
    {
    glVertex3f(game->map->width/2. + pc_cos[i]*lng*4.,
               game->sea_level-game->sea_variations,
	       game->map->height/2. + pc_sin[i]*lng*4.);
    glVertex3f(game->map->width/2., -5., game->map->height/2.);
    glVertex3f(game->map->width/2. + pc_cos[i+1]*lng*4.,
               game->sea_level-game->sea_variations,
	       game->map->height/2. + pc_sin[i+1]*lng*4.);
    }
  glEnd();
*/
  glBegin(GL_QUADS);

  /* North face */
  oldw = 0;
  theta = atan((oldw - half_width) / half_height);
  old_sin_theta = sin(theta);
  old_cos_theta = cos(theta);
  for(w = 0.125 * width; w <= width; w += 0.125 * width)
  {
    theta = atan((w - half_width) / half_height);
    sin_theta = sin(theta);
    cos_theta = cos(theta);

    glVertex3f(oldw, 0.0, 0.0);
    glVertex3f(w, 0.0, 0.0);
    glVertex3f(half_width + radius * sin_theta, 0.0, half_height + radius * -cos_theta);
    glVertex3f(half_width + radius * old_sin_theta, 0.0, half_height + radius * -old_cos_theta);

    oldw = w;
    old_sin_theta = sin_theta;
    old_cos_theta = cos_theta;
  }

  /* East face */
  oldh = 0.0;
  theta = atan((oldh - half_height) / half_height);
  old_sin_theta = sin(theta);
  old_cos_theta = cos(theta);
  for(h = 0.125 * height; h <= height; h += 0.125 * height)
  {
    theta = atan((h - half_height) / half_height);
    sin_theta = sin(theta);
    cos_theta = cos(theta);

    glVertex3f(width, 0.0, oldh);
    glVertex3f(width, 0.0, h);
    glVertex3f(half_width + radius * cos_theta, 0.0, half_height + radius * sin_theta);
    glVertex3f(half_width + radius * old_cos_theta, 0.0, half_height + radius * old_sin_theta);

    oldh = h;
    old_sin_theta = sin_theta;
    old_cos_theta = cos_theta;
  }

  /* South face */
  oldw = width;
  theta = -atan((oldw - half_width) / half_height);
  old_sin_theta = sin(theta);
  old_cos_theta = cos(theta);
  for(w = width - 0.125 * width; w >= 0.0; w -= 0.125 * width)
  {
    theta = -atan((w - half_width) / half_height);
    sin_theta = sin(theta);
    cos_theta = cos(theta);

    glVertex3f(oldw, 0.0, height);
    glVertex3f(w, 0.0, height);
    glVertex3f(half_width - radius * sin_theta, 0.0, half_height + radius * cos_theta);
    glVertex3f(half_width - radius * old_sin_theta, 0.0, half_height + radius * old_cos_theta);

    oldw = w;
    old_sin_theta = sin_theta;
    old_cos_theta = cos_theta;
  }

  /* West face */
  oldh = height;
  theta = -atan((oldh - half_height) / half_height);
  old_sin_theta = sin(theta);
  old_cos_theta = cos(theta);
  for(h = height - 0.125 * height; h >= 0.0; h -= 0.125 * height)
  {
    theta = -atan((h - half_height) / half_height);
    sin_theta = sin(theta);
    cos_theta = cos(theta);

    glVertex3f(0.0, 0.0, oldh);
    glVertex3f(0.0, 0.0, h);
    glVertex3f(half_height - radius * cos_theta, 0.0, half_height - radius * sin_theta);
    glVertex3f(half_height - radius * old_cos_theta, 0.0, half_height - radius * old_sin_theta);

    oldh = h;
    old_sin_theta = sin_theta;
    old_cos_theta = cos_theta;
  }

  /* Draw the trees */
  /* !!note: to be added */
  /*if(world->is_trees)
    draw_trees(world);*/

  glEnd();
}

/* Draws the ground */
static void draw_ground(Game *game)
{
  Map *map = game->map;
  Camera *camera = game->round->camera;
  int cx, cy; /* camera map location */
  int x0, x1, x2, xa, xb, xc, y0, y1, y2, ya, yb, yc; /* map boundaries */
  int xcl, xcr, ycu, ycl; /* The co-ordinates of the central square */

  unsigned int LOD1 = game->world->LOD_boundary1, LOD2 = game->world->LOD_boundary2;
  /* TEMP: trying to remove LOD3 */
  unsigned int LOD3 = 999;

  /* Lighting is precalculated */
  glPushAttrib(GL_LIGHTING_BIT);
  glDisable(GL_LIGHTING);

  /* Approx camera position on map */
  cx = (int) (camera->actual_target_pos[0] + 0.5f) / map->x_resolution;
  cy = (int) (camera->actual_target_pos[2] + 0.5f) / map->y_resolution;
  if(cx > map->size_x)
    cx = map->size_x;
  if(cy > map->size_y)
    cy = map->size_y;
  if(cx < 0)
    cx = 0;
  if(cy < 0)
    cy = 0;

  /* LOD Ranges:                          */
  /* (0,0)         y2                     */
  /*   +-----------3------------+         */
  /*   |                        |         */
  /*   |                        |         */
  /*   |           y1           |         */
  /*   |.....+-----2------+.....|         */
  /*   |     |            |     |         */
  /*   |     |     y0     |     |         */
  /*   |     |...+-1--+...|     |         */
  /*   |     |   |    |   |     |         */
  /* x2|   x1| x0|    |xa |xb   |xc       */
  /*   |     |...+----+...|     |         */
  /*   |     |     ya     |     |         */
  /*   |     |            |     |         */
  /*   |.....+------------+.....|         */
  /*   |           yb           |         */
  /*   |                        |         */
  /*   |                        |         */
  /*   +------------------------+         */
  /*               yc                     */
  /*                                      */
  /*        < Non visible map >           */

  /* Calculate distances to LOD boundaries (as in above figure) */
  x0 = cx - LOD1;
  x0 -= x0 % 2;
  if(x0 < 0)
    x0 = 0;

  x1 = x0 - LOD2;
  x1 -= x1 % 4;
  if(x1 < 0)
    x1 = 0;

  x2 = x1 - LOD3;
  if(x2 < 0)
    x2 = 0;

  xa = cx + LOD1;
  xa += xa % 2;
  if(xa >= map->size_x)
  {
    xa = map->size_x - 1;
    xa -= xa % 2;
  }

  xb = xa + LOD2;
  xb += 4 - xb % 4;
  if(xb >= map->size_x)
  {
    xb = map->size_x - 1;
    xb -= xb % 4;
  }

  xc = xb + LOD3;
  if(xc >= map->size_x)
    xc = map->size_x - 1;
  xc = xc - xc % 4;

  y0 = cy - LOD1;
  y0 -= y0 % 2;
  if(y0 < 0)
    y0 = 0;

  y1 = y0 - LOD2;
  y1 -= y1 % 4;
  if(y1 < 0)
    y1 = 0;

  y2 = y1 - LOD3;
  if(y2 < 0)
    y2 = 0;

  ya = cy + LOD1;
  ya += ya % 2;
  if(ya >= map->size_y)
  {
    ya = map->size_y - 1;
    ya -= ya % 2;
  }

  yb = ya + LOD2;
  yb += 4 - yb % 4;
  if(yb >= map->size_y)
  {
    yb = map->size_y - 1;
    yb -= yb % 4;
  }

  yc = yb + LOD3;
  if(yc >= map->size_y)
    yc = map->size_y - 1;
  yc = yc - yc % 4;

  /* First detail level 1x1 (the easy one!) */
  xcl = x0 + 1;
  xcr = xa - 1;
  ycl = ya - 1;
  ycu = y0 + 1;
  draw_map_segment(map, xcl, ycu, xcr, ycl, 1);
  draw_xjoiner(map, x0, xa, xcl, xcr, ycu, y0, 1, 2);
  draw_xjoiner_reverse(map, x0, xa, xcl, xcr, ycl, ya, 1, 2);
  draw_yjoiner(map, y0, ya, ycu, ycl, xcl, x0, 1, 2);
  draw_yjoiner_reverse(map, y0, ya, ycu, ycl, xcr, xa, 1, 2);

  /* Second detail level 2x2 */
  draw_map_loop(map, x1, y1, xb, yb, x0, y0, xa, ya, 2, 4);

  /* Third detail level 4x4 */
  draw_outer_map_loop(map, x2, y2, xc, yc, x1, y1, xb, yb, 4);

  glPopAttrib(); /* GL_LIGHTING_BIT */
}

/* Draws the water */
void draw_water(Game *game)
{
  int i;

  if(!game->world->is_water)
    return;

  glPushAttrib(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  glDepthMask(GL_FALSE);
  glDisable(GL_TEXTURE_2D);
  glEnable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

  glColor4fv(game->map->sea_colour);
  glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, game->map->sea_colour);
  glMaterialf(GL_FRONT, GL_SHININESS, 0.0);

  /*glBindTexture(GL_TEXTURE_2D, game->map->water_texture);*/

  /* YP: new sea (circular) (to do: create a GL list) */
  game->effective_sea_level = game->sea_level+game->sea_variations*cos(game->sea_frequency*game->world->total_elapsed_time);
  glNormal3f(0., 1., 0.);
  glBegin(GL_TRIANGLES);
  for(i=0; i<16; i++)
    {
    glVertex3f(game->map->width/2. + pc_cos[i]*game->map->width*4.,
               /*game->effective_sea_level,*/
               game->sea_level,
	       game->map->height/2. + pc_sin[i]*game->map->height*4.);
    glVertex3f(game->map->width/2.,
               game->effective_sea_level,
	       game->map->height/2.);
    glVertex3f(game->map->width/2. + pc_cos[i+1]*game->map->width*4.,
               /*game->effective_sea_level,*/
               game->sea_level,
	       game->map->height/2. + pc_sin[i+1]*game->map->height*4.);
    }
  glEnd();

  if(game->map->lakes)
    draw_lakes(game);

  glPopAttrib();
}

/* Draw the lakes */
static void draw_lakes(Game *game)
{
  int i = 0;

  while(game->map->lakes[i] != NULL)
  {
    glColor4fv(game->map->lake_colour);
    glCallList(game->world->lists.lakes + i);
    i++;
  }
}

/* Draw the sky */
static void draw_sky(Game *game)
{
  glPushAttrib(GL_LIGHTING_BIT);
  glDisable(GL_LIGHTING);

  glEnable(GL_TEXTURE_2D);
  glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
  glBindTexture(GL_TEXTURE_2D, game->map->sky_texture);
  
  /* YP: I change fog in order to let the sky visible */
  if(game->world->is_fog)
  {
    float lng;
    
    lng = nMAX(game->map->height, game->map->width);
    glFogf(GL_FOG_DENSITY, 0.001);
    glFogf(GL_FOG_END, lng * 1.01);
    glFogf(GL_FOG_START, lng * 0.99);
  }

  glCallList(game->world->lists.sky);
  glDisable(GL_TEXTURE_2D);
  if(game->world->is_fog)
    {
    glFogf(GL_FOG_DENSITY, game->map->fog_density);
    glFogf(GL_FOG_END, game->map->fog_end);
    glFogf(GL_FOG_START, 0.);
    }

  glPopAttrib();
}

/* Draws a rectangle of land with a rectangular hole in it */
/* NO bounds checking */
static void draw_map_loop(Map *map,
			  int xo1, int yo1,
			  int xo2, int yo2,
			  int xi1, int yi1,
			  int xi2, int yi2,
			  int stepi, int stepo)

{
  int xl, xr, yu, yl;
  /* The map loop:                       */
  /*                                     */
  /*(xo1, yo1)                           */
  /*    +-------------------------+      */
  /*    |A                        |      */
  /*    |          LAND           |      */
  /*    | (xi1, yi1)              |      */
  /*    |.....+-------------+.....|      */
  /*    |B    |             |C    |      */
  /*    |     |             |     |      */
  /*    |     |             |     |      */
  /*    |     |    HOLE     |     |      */
  /*    |     |             |     |      */
  /*    |     |             |     |      */
  /*    |     |             |     |      */
  /*    |.....+-------------+.....|      */
  /*    |D              (xi2, yi2)|      */
  /*    |                         |      */
  /*    |                         |      */
  /*    +-------------------------+      */
  /*                          (xo2, yo2) */

  /* Leave a border around the outer edge to put strips that link each step */
  /* If the hole is hard up against an edge though, don't bother -- edge of map */
  if(xo1 == xi1) /* If hole is hard up against left edge */
    xl = xo1;
  else
    xl = xo1 + stepi;
  if(xo2 == xi2) /* If hole is hard up against right edge */
    xr = xo2;
  else
    xr = xo2 - stepi;
  if(yo1 == yi1) /* If hole is hard up against top edge */
    yu = yo1;
  else
    yu = yo1 + stepi;
  if(yo2 == yi2) /* If hole is hard up against bottom edge */
    yl = yo2;
  else
    yl = yo2 - stepi;

  /* Draw the bulk of the ground */
  draw_map_segment(map, xl, yu, xr, yi1, stepi);   /* A */
  draw_map_segment(map, xl, yi1, xi1, yi2, stepi); /* B */
  draw_map_segment(map, xi2, yi1, xr, yi2, stepi); /* C */
  draw_map_segment(map, xl, yi2, xr, yl, stepi);   /* D */

  /* Draw the strips that link the edges */
  draw_xjoiner(map, xo1, xo2, xl, xr, yu, yo1, stepi, stepo);
  draw_xjoiner_reverse(map, xo1, xo2, xl, xr, yl, yo2, stepi, stepo);
  draw_yjoiner(map, yo1, yo2, yu, yl, xl, xo1, stepi, stepo);
  draw_yjoiner_reverse(map, yo1, yo2, yu, yl, xr, xo2, stepi, stepo);
}

/* Draws a rectangle of land with a rectangular hole in it as before,
   but doesn't bother to link the outer edges */
/* NO bounds checking */
static void draw_outer_map_loop(Map *map,
				int xo1, int yo1,
				int xo2, int yo2,
				int xi1, int yi1,
				int xi2, int yi2,
				int step)
{
  /* Draw the ground pieces */
  draw_map_segment(map, xo1, yo1, xo2, yi1, step); /* A */
  draw_map_segment(map, xo1, yi1, xi1, yi2, step); /* B */
  draw_map_segment(map, xi2, yi1, xo2, yi2, step); /* C */
  draw_map_segment(map, xo1, yi2, xo2, yo2, step); /* D */
}

/* Draws a map segment from (x1,y1) to (x2,y2)
 * Is designed for speed so has NO bounds checking */
static void draw_map_segment(Map *map, int x1, int y1, int x2, int y2, int step)
{
  /* draw segment */
  for(; y1 <= (y2 - step); y1 += step)
  {
    draw_xstrip(map, x1, y1, x2, y1 + step, step);
  }
}

/* draws a one width strip between two x-values, NO step checking */
static void draw_xstrip(Map *map, int x1, int y1, int x2, int y2, int step)
{
  glBegin(GL_TRIANGLE_STRIP);

  /* +-----> +x      */
  /* | ____________  */
  /* | | /| /| /| /| */
  /*\/ |/_|/_|/_|/_| */
  /*+z               */

  /* Draw segment */
  for(; x1 <= x2; x1 += step)
  {
    mapVertex(map, x1, y1);
    mapVertex(map, x1, y2);
  }

  glEnd();
}

/* Joins two segments with differing steps by a strip in the x-axis */
/* stepo should be twice stepi */
static void draw_xjoiner(Map *map,
			 int x1, int x2,
			 int xa, int xb,
			 int yi, int yo,
			 int stepi, int stepo)
{
  /* If will have no effect */
  if(yi == yo)
    return;

  /* Strip Shape:    */
  /*                 */
  /* x1       x2     */
  /* ___________  yo */
  /* \_________/     */
  /*  xa     xb   yi */

  /* Shape of triangles:                                */
  /*                                                    */
  /* stepo edge   (x1,yo)                       (x2,yo) */
  /*   +-----+       +-----+-----+-----+-----+-----+    */
  /*   |\ 2 /|       .\   /|\   /|\   /|\   /|\   /     */
  /*   |1\ /3|       . \ / | \ / | \ / | \ / | \ /      */
  /*   +--+--+       ...+--+--+--+--+--+--+--+--+       */
  /* stepi edge      (xa,yi)                 (xb,yi)    */

  glBegin(GL_TRIANGLES);

  /* If left edge is square, draw extra triangle (shown with '.') */
  if(x1 == xa)
  {
    mapVertex(map, x1, yo);
    mapVertex(map, xa, yi);
    xa += stepi;
    mapVertex(map, xa, yi);
  }
  /* Similarly on the rhs */
  if(x2 == xb)
  {
    mapVertex(map, xb, yi);
    mapVertex(map, x2, yo);
    xb -= stepi;
    mapVertex(map, xb, yi);
  }

  mapVertex(map, x1, yo);
  mapVertex(map, xa, yi);

  for(x1 += stepo; x1 < xb; x1 += stepo)
  {
    /* Finish triangle 2 */
    mapVertex(map, x1, yo);

    /* Triangle 3 */
    mapVertex(map, x1, yo);
    mapVertex(map, x1 - stepi, yi);
    mapVertex(map, x1, yi);

    /* Triangle 1 */
    mapVertex(map, x1, yo);
    mapVertex(map, x1, yi);
    mapVertex(map, x1 + stepi, yi);

    /* Start Triangle 2 */
    mapVertex(map, x1, yo);
    mapVertex(map, x1 + stepi, yi);
  }

  mapVertex(map, x2, yo);

  glEnd();
}

/* Joins two segments with differing steps by a strip in the x-axis
   (opposite to the above so triangles culled correctly) */
/* stepo should be twice stepi */
static void draw_xjoiner_reverse(Map *map,
				 int x1, int x2,
				 int xa, int xb,
				 int yi, int yo,
				 int stepi, int stepo)
{
  /* If will have no effect */
  if(yi == yo)
    return;

  /* Strip Shape:    */
  /*                 */
  /* xa       xb     */
  /*  _________   yi */
  /* /_________\     */
  /*x1         x2 yo */

  /* Shape of triangles:                                */
  /*                                                    */
  /* stepi edge      (xa,yi)                 (xb,yi)    */
  /*   +-----+       ...+--+--+--+--+--+--+--+--+       */
  /*   |3/ \1|       . / \ | / \ | / \ | / \ | / \      */
  /*   |/ 2 \|       ./   \|/   \|/   \|/   \|/   \     */
  /*   +--+--+       +-----+-----+-----+-----+-----+    */
  /* stepo edge   (x1,yo)                       (x2,yo) */

  glBegin(GL_TRIANGLES);

  /* If left edge is square, draw extra triangle (shown with '.') */
  if(x1 == xa)
  {
    mapVertex(map, xa, yi);
    mapVertex(map, x1, yo);
    xa += stepi;
    mapVertex(map, xa, yi);
  }
  /* Similarly on the rhs */
  if(x2 == xb)
  {
    mapVertex(map, x2, yo);
    mapVertex(map, xb, yi);
    xb -= stepi;
    mapVertex(map, xb, yi);
  }

  mapVertex(map, x2, yo);
  mapVertex(map, xb, yi);

  for(x2 -= stepo; x2 > xa; x2 -= stepo)
  {
    /* Finish triangle 2 */
    mapVertex(map, x2, yo);

    /* Triangle 3 */
    mapVertex(map, x2, yo);
    mapVertex(map, x2 + stepi, yi);
    mapVertex(map, x2, yi);

    /* Triangle 1 */
    mapVertex(map, x2, yo);
    mapVertex(map, x2, yi);
    mapVertex(map, x2 - stepi, yi);

    /* Start Triangle 2 */
    mapVertex(map, x2, yo);
    mapVertex(map, x2 - stepi, yi);
  }

  mapVertex(map, x1, yo);

  glEnd();
}

/* Joins two segments with differing steps by a strip in the y-axis */
/* stepo should be twice stepi */
static void draw_yjoiner(Map *map,
			 int y1, int y2,
			 int ya, int yb,
			 int xi, int xo,
			 int stepi, int stepo)
{
  /* If will have no effect */
  if(xi == xo)
    return;

  /* Strip Shape:     */
  /*                  */
  /* xo xi            */
  /*  ,   y1          */
  /*  |\              */
  /*  | | ya          */
  /*  | |             */
  /*  | |             */
  /*  | | yb          */
  /*  |/              */
  /*  `   y2          */

  /* Shape of triangles:                  */
  /*                                      */
  /*                                      */
  /*       +--+         +... (xo, y1)     */
  /*       |\1|         |\ .              */
  /* stepo | \| stepi   | \.              */
  /*  edge |2 + edge    |  + (xi, ya)     */
  /*       | /|         | /|              */
  /*       |/3|         |/ |              */
  /*       +--+         +--+              */
  /*                    |\ |              */
  /*                    | \|              */
  /*                    |  + (xi, yb)     */
  /*                    | /               */
  /*                    |/                */
  /*                    +    (xo, y2)     */
  
  glBegin(GL_TRIANGLES);

  /* If top edge is square, draw extra triangle (shown with '.') */
  if(y1 == ya)
  {
    mapVertex(map, xi, ya);
    mapVertex(map, xo, y1);
    ya += stepi;
    mapVertex(map, xi, ya);
  }
  /* Similarly on the bottom edge */
  if(y2 == yb)
  {
    mapVertex(map, xo, y2);
    mapVertex(map, xi, yb);
    yb -= stepi;
    mapVertex(map, xi, yb);
  }

  mapVertex(map, xo, y2);
  mapVertex(map, xi, yb);

  for(y2 -= stepo; y2 > ya; y2 -= stepo)
  {
    /* Finish triangle 2 */
    mapVertex(map, xo, y2);

    /* Triangle 1 */
    mapVertex(map, xo, y2);
    mapVertex(map, xi, y2 + stepi);
    mapVertex(map, xi, y2);

    /* Triangle 3 */
    mapVertex(map, xo, y2);
    mapVertex(map, xi, y2);
    mapVertex(map, xi, y2 - stepi);

    /* Start Triangle 2 */
    mapVertex(map, xo, y2);
    mapVertex(map, xi, y2 - stepi);
  }

  mapVertex(map, xo, y1);

  glEnd();
}

/* Joins two segments with differing steps by a strip in the y-axis 
   (opposite to the above so triangles culled correctly) */
/* stepo should be twice stepi */
static void draw_yjoiner_reverse(Map *map,
				 int y1, int y2,
				 int ya, int yb,
				 int xi, int xo,
				 int stepi, int stepo)
{
  /* If will have no effect */
  if(xi == xo)
    return;

  /* Strip Shape:     */
  /*                  */
  /* xi xo            */
  /*    , y1          */
  /*   /|             */
  /*  | | ya          */
  /*  | |             */
  /*  | |             */
  /*  | | yb          */
  /*   \|             */
  /*    ` y2          */

  /* Shape of triangles:                     */
  /*                                         */
  /*                                         */
  /*       +--+                ...+ (xo, y1) */
  /*       |1/|                . /|          */
  /* stepi |/ | stepo          ./ |          */
  /*  edge + 2| edge  (xi, ya) +  |          */
  /*       |\ |                |\ |          */
  /*       |3\|                | \|          */
  /*       +--+                +--+          */
  /*                           | /|          */
  /*                           |/ |          */
  /*                  (xi, yb) +  |          */
  /*                            \ |          */
  /*                             \|          */
  /*                              + (xo, y2) */
  
  glBegin(GL_TRIANGLES);

  /* If top edge is square, draw extra triangle (shown with '.') */
  if(y1 == ya)
  {
    mapVertex(map, xo, y1);
    mapVertex(map, xi, ya);
    ya += stepi;
    mapVertex(map, xi, ya);
  }
  /* Similarly on the bottom edge */
  if(y2 == yb)
  {
    mapVertex(map, xi, yb);
    mapVertex(map, xo, y2);
    yb -= stepi;
    mapVertex(map, xi, yb);
  }

  mapVertex(map, xo, y1);
  mapVertex(map, xi, ya);

  for(y1 += stepo; y1 < yb; y1 += stepo)
  {
    /* Finish triangle 2 */
    mapVertex(map, xo, y1);

    /* Triangle 3 */
    mapVertex(map, xo, y1);
    mapVertex(map, xi, y1 - stepi);
    mapVertex(map, xi, y1);

    /* Triangle 1 */
    mapVertex(map, xo, y1);
    mapVertex(map, xi, y1);
    mapVertex(map, xi, y1 + stepi);

    /* Start Triangle 2 */
    mapVertex(map, xo, y1);
    mapVertex(map, xi, y1 + stepi);
  }

  mapVertex(map, xo, y2);

  glEnd();
}

/* Draws a vertex from the map at (x,y) */
static void mapVertex(Map *map, int x, int y)
{
  float *p, c[3];

  p = map->pos[x][y];

  c[0] = map->static_light[x][y][0] + map->dynamic_light[x][y][0];
  c[1] = map->static_light[x][y][1] + map->dynamic_light[x][y][1];
  c[2] = map->static_light[x][y][2] + map->dynamic_light[x][y][2];

  glColor3fv(c);
  glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, c);
  glNormal3fv(map->norm[x][y]);
  glTexCoord2f((GLfloat)x * map->tex_scale_x, (GLfloat)y * map->tex_scale_y);

  /*glVertex3f((float)x * map->x_resolution, map->pos[x][y][1], (float)y * map->y_resolution);*/
  glVertex3fv(p);
}

/* Same as above but h is calculated to keep the accuracy the same with step */
static void mapVertexs(Map *map, int x, int y, int step)
{
  float h, h0, h1, h2, stepf = (float)step, c[3];
  float *p;
  int xr = x % step, yr = y % step, xc = x - xr, yc = y - yr;

  /*     <-step-->      s=step    */
  /*     <-xr->                   */
  /*   h1+-+-+-+-+h0'=(xc+s,yc+s) */
  /*     |\| | | |                */
  /*     +-+-+-+-+                */
  /*     | |\| | |                */
  /*     +-+-+-p-+ p=point@(x,y)  */
  /*   /\| | |\| |                */
  /*   yr+-+-+-+-+                */
  /*   \/| | | |\|                */
  /*   h0+-+-+-+-+h2              */
  /* (xc,yc)                      */

  p = map->pos[x][y];

  h1 = map->pos[xc][yc + step][1];
  h2 = map->pos[xc + step][yc][1];

  if((xr + yr) > step) /* upper triangle */
  {
    /* calculate height */
    h0 = map->pos[xc + step][yc + step][1];
    h = h0 + (((float)(step - xr) / stepf)) * (h1 - h0)
           + (((float)(step - yr) / stepf)) * (h2 - h0);
  } 
  else /* lower traingle */
  {
    /* calculate height */
    h0 = map->pos[xc][yc][1];
    h = h0 + (((float)xr / stepf)) * (h2 - h0)
           + (((float)yr / stepf)) * (h1 - h0);
  }

  c[0] = map->static_light[x][y][0] + map->dynamic_light[x][y][0];
  c[1] = map->static_light[x][y][1] + map->dynamic_light[x][y][1];
  c[2] = map->static_light[x][y][2] + map->dynamic_light[x][y][2];

  glColor3fv(c);
  glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, c);
  glNormal3fv(map->norm[x][y]); /* !!note: work here? */
  glTexCoord2f((GLfloat)x * map->tex_scale_x, (GLfloat)y * map->tex_scale_y);

  glVertex3fv(p);
}
