/* 
 * NATOW - Tank to tank warfare in a 3D fully destructable world
 *
 * Copyright (C) 2000-2001
 *                     Robert Ancell  <bob27@users.sourceforge.net>
 *                     Yannick Perret <hexasoft@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* game.c
   Start and stop games */

#include <string.h>

#include "game.h"
#include "world.h"
#include "text.h"
#include "2d.h"
#include "ai.h"
#include "player.h"
#include "player_list.h"
#include "draw.h"
#include "tank.h"
#include "tanks.h"
#include "munition.h"
#include "satellite.h"
#include "level.h"
#include "map.h"
#include "videoc.h"
#include "file.h"
#include "input.h"
#include "image.h"
#include "bird.h"
#include "tools.h"
#include "update.h"
#include "cursor.h"
#include "glut_interface.h"
#include "display_stack.h"
#include "light.h"

static Game *allocate_game(void);

/* Make an empty game structure */
static Game *allocate_game(void)
{
  Game *game;

  game = malloc(sizeof(Game));
  if(!game)
  {
    Neprintf("Error: Unable to allocate space for a new game\n");
    return NULL;
  }

  game->world = NULL;

  game->level = NULL;
  game->map = NULL;

  /* The list of objects */
  game->nobjects = 0;
  game->object_list = NULL;

  game->mode = TURN_BASED;

  game->gravity = -9.81;
  game->sea_level = 10.0;
  game->sea_variations = 1.0;
  game->sea_frequency = 0.2;
  game->effective_sea_level = 10.0;

  game->starting_cash = 0;
  game->driving_time = 0.0;

  game->players = NULL;

  game->round_num = 0;
  game->max_rounds = 3;
  game->round = NULL;

  return game;
}

/* Create a game */
int create_game(World *world)
{
  Game *game;

  if(!world)
    return FALSE;

  if(world->game)
  {
    Neprintf("Error: Trying to start game, but game already started!\n");
    return FALSE;
  }

  game = world->game = allocate_game();
  if(!game)
    return FALSE;

  game->players = new_player_list();
  if(!game->players)
  {
    free_game(game);
    return FALSE;
  }

  /* Choose the intial game mode */
  set_game_mode(game, TURN_BASED);

  game->world = world;

  game->round_num = 0;

  Nprintf("Game created\n");

  return TRUE;
}

/* Set the game mode, i.e. turn-based/real-time/single-player */
int set_game_mode(Game *game, GameMode mode)
{
  World *world;

  if(!game || !game->world)
    return FALSE;

  world = game->world;
  game->mode = mode;

  /* Default: unused */
  world->display_func = NULL;
  world->reshape_func = NULL;
  world->keyboard_func = NULL;
  world->mouse_func = NULL;
  world->motion_func = NULL;
  world->update_func = NULL;
  world->move_func = NULL;
  world->entry_func = NULL;

  switch(mode)
  {
  case TURN_BASED:
    world->display_func = display;
    world->reshape_func = reshape;
    world->keyboard_func = keyboard;
    world->motion_func = motion;
    world->update_func = update_world;
    world->move_func = manage_mouse_cursor;
    world->mouse_func = mouse_event;
    world->entry_func = enter_leave;
    break;
  case REAL_TIME:
    /* FILL ME!! */
    break;
  case SINGLE_PLAYER:
    /* FIXME: These are probably not correct */
    world->display_func = display;
    world->reshape_func = reshape;
    world->keyboard_func = keyboard;
    world->motion_func = motion;
    world->update_func = update_world;
    world->move_func = manage_mouse_cursor;
    world->mouse_func = mouse_event;
    world->entry_func = enter_leave;
    break;
  default:
    Neprintf("Error: Unknown game mode (%d)\n", mode);
    break;
  }

  /* Connect these functions up */
  update_glut_functions(world);

  return TRUE;
}

/* Set the level used in this game */
int set_game_level(Game *game, LevelEntry *level)
{
  if(!game)
    return FALSE;

  game->level = level;

  return TRUE;
}

/* Set the gravity for a game */
int set_game_gravity(Game *game, float gravity)
{
  if(!game)
    return FALSE;

  game->gravity = gravity;

  return TRUE;
}

/* Set the driving time for a game */
int set_game_driving_time(Game *game, float driving_time)
{
  if(!game)
    return FALSE;

  game->driving_time = driving_time;

  return TRUE;
}

/* Set the number of rounds in a game */
int set_game_rounds(Game *game, int max_rounds)
{
  if(!game)
    return FALSE;

  game->max_rounds = max_rounds;

  return TRUE;
}

/* Set the starting cash in a game */
int set_game_starting_cash(Game *game, int starting_cash)
{
  if(!game)
    return FALSE;

  game->starting_cash = starting_cash;

  return TRUE;
}

/* Start a game */
int start_game(Game *game)
{
  if(!game)
    return FALSE;

  /* Game already started... TODO: Do this a better way? */
  if(game->map)
  {
    set_display_mode(game->world, NATOW_DISPLAY_START_ROUND);
    return TRUE;
  }

  if(!game->level)
  {
    Neprintf("Unable to start game, no level chosed\n");
    set_display_mode(game->world, NATOW_DISPLAY_MAIN_MENU);
    return FALSE;
  }

  /* As loading the level can be long and display
     is freezed during this operation, just display
     "loading" to be user friendly :o) */
  activate_2d();
  draw_2d_text_center("Loading...", 1, 98, 50, 8, 2);
  deactivate_2d();
  swap_buffers(); /* display now */

  /* Load the map etc */
  if(!load_level(game, game->level->file_name))
  {
    Neprintf("  No map loaded.\n");
    finish_game(game);
    return FALSE;
  }

  if(game->world->is_fog)
    glEnable(GL_FOG);
  glFogi(GL_FOG_MODE, GL_EXP2);
  glFogf(GL_FOG_DENSITY, game->map->fog_density);
  glFogf(GL_FOG_END, game->map->fog_end);
  glFogfv(GL_FOG_COLOR, game->map->fog_colour);

  glMatrixMode(GL_MODELVIEW);
  
  set_display_mode(game->world, NATOW_DISPLAY_START_ROUND);

  return TRUE;
}

/* Add a player to the game */
int add_player_to_game(Game *game, Player *player)
{
  if(!game || !game->players)
    return FALSE;

  append_player(game->players, player);

  return TRUE;
}

/* End a game */
int end_game(Game *game)
{
  /* Caclulate round winners */
  Nprintf("Caclulating game winners\n");
  sort_player_list(game->players, compare_game_winners);

  /* Display the results */
  set_display_mode(game->world, NATOW_DISPLAY_END_GAME);

  return TRUE;
}

/* Comparison function to compare entries for sorting */
int compare_game_winners(PlayerEntry **entry1, PlayerEntry **entry2)
{
  Player *player1, *player2;
  int t;

  player1 = (*entry1)->player;
  player2 = (*entry2)->player;

  /* Winner is the one with the most kills */
  if((t = player2->kills - player1->kills))
    return t;

  /* Winner is the one with the most money left */
  if((t = player2->cash - player1->cash))
    return t;

  /* A draw */
  return 0;
}

/* Remove everything from the game */
int finish_game(Game *game)
{
  if(!game)
    return FALSE;

  /* Go back to the main menu */
  set_display_mode(game->world, NATOW_DISPLAY_MAIN_MENU);

  game->world->game = NULL;
  free_game(game);

  Nprintf("Game finished\n");

  return TRUE;
}

/* Update the game */
void update_game(Game *game, float dt)
{
  if(game->round)
    update_round(game->round, dt);
}

/* Free the game structure and all below it */
int free_game(Game *game)
{
  if(!game)
    return FALSE;

  free_round(game->round);

  free_map(game->map);
  /* TODO */
  /*free_objects(game->object_list);*/
  free_player_list(game->players);

  Nfree(game);

  return TRUE;
}

