/* 
 * NATOW - Tank to tank warfare in a 3D fully destructable world
 *
 * Copyright (C) 2000-2001
 *                     Robert Ancell  <bob27@users.sourceforge.net>
 *                     Yannick Perret <hexasoft@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* glut_interface.c
   These functions interface n.a.t.o.w with glut
   Hopefully if glut gets thrown out, the changes will all be confined to here */

#include <stdio.h>
#include <string.h>
#include <GL/glut.h>

#include "glut_interface.h"
#include "world.h"

static World *glut_world = NULL;

/* Initialize glut */
void init_glut(int *argc, char **argv)
{
  glutInit(argc, argv);
}

/* Make a window for natow */
void make_glut_window(int w, int h)
{
  glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
  glutInitWindowSize(w, h);
  glutInitWindowPosition(40, 40);
  glutCreateWindow("n.a.t.o.w version " NATOW_VERSION_STRING);
  glutSetCursor(GLUT_CURSOR_NONE);
}

/* Reshape the window */
void reshape_glut_window(int w, int h)
{
  glutReshapeWindow(w, h);
}

/* Go to fullscreen mode */
void fullscreen_glut(void)
{
  glutFullScreen();
}

/* Move the window */
void move_glut_window(int x, int y)
{
  glutPositionWindow(x, y);
}

/* Set the world for glut (can't pass it to any of glut's functions) */
void set_glut_world(World *world)
{
  glut_world = world;
}

/* Attach the required functions to glut */
void update_glut_functions(World *world)
{
  /* Output */
  if(world->display_func)
    glutDisplayFunc(glut_display);
  else
    glutDisplayFunc(NULL);
  if(world->reshape_func)
    glutReshapeFunc(glut_reshape);
  else
    glutReshapeFunc(NULL);

  /* Input */
  if(world->keyboard_func)
  {
    glutSpecialFunc(glut_ext_keyboard);
    glutKeyboardFunc(glut_keyboard);
  }
  else
  {
    glutSpecialFunc(NULL);
    glutKeyboardFunc(NULL);
  }
  if(world->mouse_func)
    glutMouseFunc(glut_mouse);
  else
    glutMouseFunc(NULL);
  if(world->motion_func)
  {
    glutMotionFunc(glut_motion);
    /* Always use the mouse function, so you can calc. dx/dy */
    glutMouseFunc(glut_mouse);
  }
  else
    glutMotionFunc(NULL);

  /* Mouse movements */
  if(world->move_func)
    glutPassiveMotionFunc(glut_move);
  else
    glutPassiveMotionFunc(NULL);

  /* Enter/leave events */
  if(world->entry_func)
    glutEntryFunc(glut_enter_leave);
  else
    glutEntryFunc(NULL);

  /* Time */
  if(world->update_func)
    glutIdleFunc(glut_idle);
  else
    glutIdleFunc(NULL);
}

/* Whenever the window needs to be redrawn */
void glut_display(void)
{
  glut_world->display_func(glut_world);
}

/* Whenever the window is resized */
void glut_reshape(int w, int h)
{
  glut_world->reshape_func(glut_world, w, h);
}

/* The function that gets called for each extended keyboard event
   (for example arrow keys) */
void glut_ext_keyboard(int key, int x, int y)
{
  /* Add 1000 so keycodes don't overlap with basic keycodes */
  glut_world->keyboard_func(glut_world, key + 1000, x, y);
}

/* The function that gets called for each keyboard event */
void glut_keyboard(unsigned char key, int x, int y)
{
  glut_world->keyboard_func(glut_world, (int)key, x, y);
}

/* The function that gets called for each mouse movements */
void glut_move(int x, int y)
{
  glut_world->move_func(glut_world, x, y);
}

/* The function that gets called for each enter/leave for the window */
void glut_enter_leave(int state)
{
  glut_world->entry_func(glut_world, state);
}

static int mouse_oldx, mouse_oldy;
static int button_states[3] = {MOUSE_UP, MOUSE_UP, MOUSE_UP};

/* The function that gets called for each mouse click event */
void glut_mouse(int button, int state, int x, int y)
{
  mouse_oldx = x;
  mouse_oldy = y;
  if(button == GLUT_LEFT_BUTTON)
    button_states[LEFT_BUTTON] = (state == GLUT_UP);
  else if(button == GLUT_MIDDLE_BUTTON)
    button_states[MIDDLE_BUTTON] = (state == GLUT_UP);
  else if(button == GLUT_RIGHT_BUTTON)
    button_states[RIGHT_BUTTON] = (state == GLUT_UP);

  if(glut_world->mouse_func)
    glut_world->mouse_func(glut_world, button, state, x, y);
}

/* The function that gets called for each mouse motion event (while a button is pressed) */
void glut_motion(int x, int y)
{
  glut_world->motion_func(glut_world, x, y, x - mouse_oldx, y - mouse_oldy);
  mouse_oldx = x;
  mouse_oldy = y;
}

/* Get the current states of the buttons
   returns 3 entry array of states for each button */
int *get_button_states(void)
{
  return button_states;
}

/* The function that gets called for each idle event */
void glut_idle(void)
{
  float dt;
  int ticks, elapsed;
  static int last_ticks = -1;

  if(last_ticks == -1)
    last_ticks = glutGet(GLUT_ELAPSED_TIME);

  ticks = glutGet(GLUT_ELAPSED_TIME);
  elapsed = ticks - last_ticks;
  last_ticks = ticks;
  dt = (float) elapsed / 1000.0;

  glut_world->update_func(glut_world, dt);
}

/* Start glut */
void start_glut(void)
{
  glutMainLoop();
}

/* The below prints strings using GLUT (not used) */

/* Print a string */
void print_string(int x, int y, char *string )
{
  int len, i;

  glColor4f(0.0f, 0.0f, 1.0f, 0.5f);
  glRasterPos2i(x, y);
  len = (int) strlen(string);
  for (i=0; i<len; i++)
    glutBitmapCharacter(GLUT_BITMAP_HELVETICA_18, string[i]);
}

/* Print a string (let user choose the colour and transp.) */
void print_cstring(int x, int y, float r, float g, float b, float t, char *string)
{
  int len, i;

  glColor4f(r, g, b, t);
  glRasterPos2i(x, y);
  len = (int) strlen(string);
  for (i=0; i<len; i++)
    glutBitmapCharacter(GLUT_BITMAP_HELVETICA_18, string[i]);
}
