/* 
 * NATOW - Tank to tank warfare in a 3D fully destructable world
 *
 * Copyright (C) 2000-2001
 *                     Robert Ancell  <bob27@users.sourceforge.net>
 *                     Yannick Perret <hexasoft@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* light.c
   This file contains all the code for lighting the map */

#include "light.h"
#include "text.h"
#include "vector.h"
#include "map.h"

#define PC_RADIUS 100
#define PC_DIAMETER (PC_RADIUS + PC_RADIUS + 2)
static float volume_light[PC_DIAMETER][PC_DIAMETER][PC_DIAMETER];

static void calc_static_lighting(Map *map,
				 float *light_dir, float *diffuse_colour, float *ambient_colour);
static void calc_static_light(Map *map, int x, int y,
			      float *light_dir, float *light_diffuse, float *light_ambient);
static void shade_point(Map *map, int x, int y,
			float *light_dir, float *light_diffuse, float *light_ambient);
static void clear_dynamic_light_map_area(Map *map, int x0, int y0, int x1, int y1);

/* Initialize the precalculated lighting */
int init_lighting(Map *map)
{
  int i, j, k;
  float l[3];

  Nprintf("  Precalculating volume lighting...\n");

  for(i = 0; i < PC_DIAMETER; i++)
  {
    l[0] = (float)(i - PC_RADIUS) * map->x_resolution;
    for(j = 0; j < PC_DIAMETER; j++)
    {
      l[1] = (float)(j - PC_RADIUS);
      for(k = 0; k < PC_DIAMETER; k++)
      {
	l[2] = (float)(k - PC_RADIUS) * map->y_resolution;

	volume_light[i][j][k] = 1.0 / (l[0]*l[0] + l[1]*l[1] + l[2]*l[2]);
      }
    }
  }

  Nprintf("  Lighting precalculated\n");

  return TRUE;
}

/* Calculates the effect of the sun on the ground */
void calc_map_lighting(Map *map)
{
  calc_static_lighting(map, map->sun_dir, map->sun_diffuse, map->sun_ambient);
}

/* Calculates the lighting on the whole map from a light (eg the sun) */
static void calc_static_lighting(Map *map,
				 float *light_dir,
				 float *diffuse_colour,
				 float *ambient_colour)
{
  int x, y;

  if(!map)
    return;

  Nprintf("Calculating static lighting...\r");
  fflush(stdout);

  for(x = 0; x < map->size_x; x++)
  {
    Nprintf("Calculating static lighting... [%d/%d]\r", x+1, map->size_x);
    fflush(stdout);

    for(y = 0; y < map->size_y; y++)
      calc_static_light(map, x, y, light_dir, diffuse_colour, ambient_colour);
  }

  Nprintf("Calculating static lighting... [done]          \n");
  fflush(stdout);
}

/* Calculates the static light colour at the point (x, y) */
static void calc_static_light(Map *map,
			      int x, int y,
			      float *light_dir,
			      float *light_diffuse,
			      float *light_ambient)
{
  float dir[3], *p, pos[3], pos0[3], pos1[3], d[3];
  float error = 0.05;

  _vector_scale(10000.0, light_dir, dir);
  _make_vector((float)(x * map->x_resolution),
	       map->pos[x][y][1],
	       (float)(y * map->y_resolution),
	       pos);
  _vector_add(pos, dir, pos0);
  _vector_sub(pos, dir, pos1);

  /* The collision point of the ray */
  p = get_map_intersection_long(map, pos0, pos1, -1.0);

  if(p != NULL)
  {
    /* Get the difference between to ground point and the ray intersection,
       to check if that point is visible to the ray */
    _vector_sub(p, pos, d);
    if(vector_2norm(d) < error) /* Visible to the light */
      shade_point(map, x, y, light_dir, light_diffuse, light_ambient);
    else /* In shadow */
    {
      map->static_light[x][y][0] = light_ambient[0];
      map->static_light[x][y][1] = light_ambient[1];
      map->static_light[x][y][2] = light_ambient[2];
    }
  }
  /* No collision -- something is probably wrong but lets just 
     assume that the surface is visible */
  else
    shade_point(map, x, y, light_dir, light_diffuse, light_ambient);
}

/* Shade the whole map */
void shade_map(Map *map)
{
  shade_map_area(map, 0, 0, map->size_x - 1, map->size_y - 1,
		 map->sun_dir, map->sun_diffuse, map->sun_ambient);
}

/* Shades an area of the world map */
void shade_map_area(Map *map,
		    int x0, int y0, int x1, int y1,
		    float *light_dir,
		    float *light_diffuse, float *light_ambient)
{
  int x, y;

  if(x0 < 0)
    x0 = 0;
  if(y0 < 0)
    y0 = 0;
  if(x1 >= map->size_x)
    x1 = map->size_x - 1;
  if(y1 >= map->size_y)
    y1 = map->size_y - 1;

  for(x = x0; x < x1; x++)
    for(y = y0; y < y1; y++)
      shade_point(map, x, y, light_dir, light_diffuse, light_ambient);
}

/* Same as above but use the worlds sun */
void shade_map_area_with_sun(Map *map, int x0, int y0, int x1, int y1)
{
  shade_map_area(map,
		 x0, y0, x1, y1,
		 map->sun_dir,
		 map->sun_diffuse, map->sun_ambient);
}

/* Shades the map point based on a directional light */
static void shade_point(Map *map,
			int x, int y,
			float *light_dir,
			float *light_diffuse, float *light_ambient)
{
  float *n, intensity;

  n = map->norm[x][y];
  intensity = n[0]*light_dir[0] + n[1]*light_dir[1] + n[2]*light_dir[2];
  if(intensity < 0.0)
    intensity = 0.0;

  map->static_light[x][y][0] = intensity * light_diffuse[0] + light_ambient[0];
  map->static_light[x][y][1] = intensity * light_diffuse[1] + light_ambient[1];
  map->static_light[x][y][2] = intensity * light_diffuse[2] + light_ambient[2];
}

/* Add a light into the dynamic light map */
void add_dynamic_light(Map *map, float *pos, float *col, float intensity)
{
  int x, y, rx, ry, x0, y0, a, b, c, d;
  float r, dx, dy, dz, lpercent, v[3], *n, dotp, *p, *dc;
  /* Used in old method */
  /*float dist;*/

  x0 = (int)(pos[0] / map->x_resolution + 0.5);
  y0 = (int)(pos[2] / map->y_resolution + 0.5);

  /* Radius covers area with 95% of light intensity */
  r = sqrt(500.0 * intensity) + 1.0;
  /* Cannot be greater than PC_RADIUS (the precalculated radius) */
  if(r > PC_RADIUS)
    r = PC_RADIUS;
  rx = (int)(r / map->x_resolution);
  ry = (int)(r / map->y_resolution);
  /* Square approximation to circle */
  a = x0 - rx;
  b = x0 + rx;
  c = y0 - ry;
  d = y0 + ry;

  /* Clip area to map limits */
  if(a < 0)
    a = 0;
  else if(a > map->size_x)
    a = map->size_x;
  if(b < 0)
    b = 0;
  else if(b > map->size_x)
    b = map->size_x;
  if(c < 0)
    c = 0;
  else if(c > map->size_y)
    c = map->size_y;
  if(d < 0)
    d = 0;
  else if(d > map->size_y)
    d = map->size_y;

  /* Shade all the points in that square using
     precalculated values */
  for(x = a; x < b; x++)
    for(y = c; y < d; y++)
    {
      p = map->pos[x][y];
      v[0] = dx = pos[0] - p[0];
      v[1] = dy = pos[1] - p[1];
      v[2] = dz = pos[2] - p[2];

      if(fabs(dy) < r)
      {
	normalize_vector(v);
	n = map->norm[x][y];
	dotp = n[0]*v[0] + n[1]*v[1] + n[2]*v[2];

	if(dotp > 0.0)
	{
	  lpercent = dotp * intensity *
	    volume_light[x - x0 + PC_RADIUS]
	                [(int)(dy + 0.5) + PC_RADIUS]
	                [y - y0 + PC_RADIUS];

	  dc = map->dynamic_light[x][y];
	  dc[0] += lpercent * col[0];
	  dc[1] += lpercent * col[1];
	  dc[2] += lpercent * col[2];
	}
      }
    }

  /* The old method, calculate each point */
  /* Shade all the points in that square */
  /*for(x = a; x < b; x++)
    for(y = c; y < d; y++)
    {
      p = map->pos[x][y];
      v[0] = dx = pos[0] - p[0];
      v[1] = dy = pos[1] - p[1];
      v[2] = dz = pos[2] - p[2];

      normalize_vector(v);
      n = map->norm[x][y];
      dotp = n[0]*v[0] + n[1]*v[1] + n[2]*v[2];

      if(dotp > 0.0)
      {
	dist = (dx*dx + dy*dy + dz*dz);
	lpercent = dotp * intensity / dist;

	dc = map->dynamic_light[x][y];
	dc[0] += lpercent * col[0];
	dc[1] += lpercent * col[1];
	dc[2] += lpercent * col[2];
      }
    }*/
}

/* Clear all the dynamic lighting */
void clear_dynamic_light_map(Map *map)
{
  clear_dynamic_light_map_area(map, 0, map->size_x - 1, 0, map->size_y - 1);
}

/* Set the light in the rectangle (x0, y0)-(x1, y1) to black */
static void clear_dynamic_light_map_area(Map *map, int x0, int y0, int x1, int y1)
{
  int x, y;

  for(x = x0; x < x1; x++)
    for(y = y0; y < y1; y++)
    {
      map->dynamic_light[x][y][0] = 0.0;
      map->dynamic_light[x][y][1] = 0.0;
      map->dynamic_light[x][y][2] = 0.0;
    }
}

/* Clear the area affected by this dynamic light */
void clear_dynamic_light(Map *map, float *pos, float intensity)
{
  int rx, ry, x0, y0, a, b, c, d;
  float r;

  x0 = (int)(pos[0] / map->x_resolution + 0.5);
  y0 = (int)(pos[2] / map->y_resolution + 0.5);

  /* Radius covers area with 95% of light intensity */
  r = sqrt(500.0 * intensity) + 1.0;
  rx = (int)(r / map->x_resolution);
  ry = (int)(r / map->y_resolution);
  a = x0 - rx;
  b = x0 + rx;
  c = y0 - ry;
  d = y0 + ry;

  if(a < 0)
    a = 0;
  else if(a > map->size_x)
    a = map->size_x;
  if(b < 0)
    b = 0;
  else if(b > map->size_x)
    b = map->size_x;
  if(c < 0)
    c = 0;
  else if(c > map->size_y)
    c = map->size_y;
  if(d < 0)
    d = 0;
  else if(d > map->size_y)
    d = map->size_y;

  clear_dynamic_light_map_area(map, a, c, b, d);
}
