/* 
 * NATOW - Tank to tank warfare in a 3D fully destructable world
 *
 * Copyright (C) 2000-2001
 *                     Robert Ancell  <bob27@users.sourceforge.net>
 *                     Yannick Perret <hexasoft@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef __map_h__
#define __map_h__

#include <GL/gl.h>

typedef struct _Map Map;
typedef struct _StartingPoint StartingPoint;

#include "world.h"
#include "texture.h"
#include "lakes.h"
#include "objects.h"

/* The map to fight over */
struct _Map
{
  Game *game;

  /* The name of the map */
  char *name;

  /* map properties */
  float x_resolution, y_resolution;
  unsigned int size_x, size_y;

  /* NULL terminated array of starting points */
  StartingPoint **startingpoints;

  float width, height, max_altitude;

  /* The map (surface) texture */
  /* The size of the texture (calculated from above) */
  int tex_width, tex_height;
  /* The actual texture data */
  GLubyte *texture_data;
  /* The index of the openGL texture */
  GLuint texture;
  /* The scale to multiply each vertex by to get the tex co-ord. */
  GLfloat tex_scale_x, tex_scale_y;

  /* The sun properties */
  GLfloat sun_dir[4], sun_diffuse[4], sun_ambient[4];

  /* The mini-textures that make up the map texture */
  char *rock_tex_fname, *ice_tex_fname, *tree_tex_fname;
  char *grass_tex_fname, *sand_tex_fname, *undersea_tex_fname;
  NatowTexture *rock_texture;
  NatowTexture *ice_texture;
  NatowTexture *tree_texture;
  NatowTexture *grass_texture;
  NatowTexture *sand_texture;
  NatowTexture *undersea_texture;

  /* YP: the sky texture */
  GLuint sky_texture;

  /* A 2d array with a vector that is the actual map point
     from the grid */
  float ***pos;
  /* A 2d array containing the normals of the
     map at grid co-ordinates */
  float ***norm;

  /* objects map */
  struct _Object ***objects;

  /* YP: for lake support */
  Lake **lakes;  /* NULL terminated list of lakes */
  GLfloat lake_colour[4];

  GLfloat sea_colour[4];

  /* The static and dynamic light maps */
  float ***static_light, ***dynamic_light;

  /* number of birds requested */
  int birds_number;

  /* The fog color used and other fog properties */
  GLfloat fog_colour[4];
  GLfloat fog_density;
  GLfloat fog_end;
};

/* Starting points for players */
struct _StartingPoint
{
  float x, y;
  int selected;
};

/* Free map (and associated) arrays */
int free_map(Map *map);

/* Loads a map file */
Map *load_map(char *fname, float max_altitude);

/* Randomizes a map */
void randomize_map(Map *map);

/* Calculates the normals for the whole map */
void calc_normals(Map *map);

/* Calculates the map normal at the co-ordinates (x, y) */
void calc_normal(Map *map, int x, int y);

/* Gets the intersection when there's a large distance between the two points */
float *get_map_intersection_long(Map *map, float *posa, float *posb, float error);

/* Returns the point of intersection of a line and the map */
float *get_map_intersection(Map *map, float *posa, float *posb, float error);

/* Gets the height of the map at an arbitrary co-ordinate (x, y) */
float get_height(Map *map, float x, float y);

/* Gets a angular normal at an arbitrary co-ordinate (x, y),
   (Value is used internally -- do not free) */
float *get_normal(Map *map, float x, float y);

/* Gets a smoothed normal at an arbitrary co-ordinate (x, y),
   (Value is used internally -- do not free) */
float *get_smooth_normal(Map *map, float x, float y);

/* Gets the gradient of the map at (x,y) */
float *get_gradient(Map *map, float x, float y);

/* Get the rotation required in the x-axis to put an object normal to
 * the ground */
float get_rotx(Map *map, float x, float y);

/* Same as above but in the z-axis */
float get_rotz(Map *map, float x, float y);

/* Test whether a point is in the bounds of the map or not,
   returning TRUE if in bounds, FALSE otherwise */
int in_bounds(Map *map, float x, float y);

/* Same as above, but for a 3D vector */
int in_boundsv(Map *map, float *pos);

#endif /* __map_h__ */
