/* 
 * NATOW - Tank to tank warfare in a 3D fully destructable world
 *
 * Copyright (C) 2000-2001
 *                     Robert Ancell  <bob27@users.sourceforge.net>
 *                     Yannick Perret <hexasoft@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* map_texture.c
   Contains the functions to generate the texture on the map */

#include <math.h>

#include "map_texture.h"
#include "world.h"
#include "game.h"
#include "text.h"
#include "texture.h"
#include "map.h"
#include "tools.h"

/* Load all the component textures required to make the whole map texture,
   return TRUE if all textures loaded */
int load_mini_textures(Map *map)
{
  /* Load the rock texture */
  if(map->rock_texture == NULL && map->rock_tex_fname != NULL)
  {
    Nprintf("    Loading rock texture file: \"%s\"\n", map->rock_tex_fname);
    map->rock_texture = load_natow_texture(map->rock_tex_fname);
    if(map->rock_texture == NULL)
      Neprintf("      Error loading rock texture file: \"%s\"! (skipped)\n",
	       map->rock_tex_fname);
  }

  /* Load the ice texture */
  if(map->ice_texture == NULL && map->ice_tex_fname != NULL)
  {
    Nprintf("    Loading ice texture file: \"%s\"\n", map->ice_tex_fname);
    map->ice_texture = load_natow_texture(map->ice_tex_fname);
    if(map->ice_texture == NULL)
      Neprintf("      Error loading ice texture file: \"%s\"! (skipped)\n",
	       map->ice_tex_fname);
  }

  /* Load the (under) tree texture */
  if(map->tree_texture == NULL && map->tree_tex_fname != NULL)
  {
    Nprintf("    Loading tree texture file: \"%s\"\n", map->tree_tex_fname);
    map->tree_texture = load_natow_texture(map->tree_tex_fname);
    if(map->tree_texture == NULL)
      Neprintf("      Error loading tree texture file: \"%s\"! (skipped)\n",
	       map->tree_tex_fname);
  }

  /* Load the grass texture */
  if(map->grass_texture == NULL && map->grass_tex_fname != NULL)
  {
    Nprintf("    Loading grass texture file: \"%s\"\n", map->grass_tex_fname);
    map->grass_texture = load_natow_texture(map->grass_tex_fname);
    if(map->grass_texture == NULL)
      Neprintf("      Error loading grass texture file: \"%s\"! (skipped)\n",
	       map->grass_tex_fname);
  }

  /* Load the sand texture */
  if(map->sand_texture == NULL && map->sand_tex_fname != NULL)
  {
    Nprintf("    Loading sand texture file: \"%s\"\n", map->sand_tex_fname);
    map->sand_texture = load_natow_texture(map->sand_tex_fname);
    if(map->sand_texture == NULL)
      Neprintf("      Error loading sand texture file: \"%s\"! (skipped)\n",
	       map->sand_tex_fname);
  }

  /* Load the under-sea texture */
  if(map->undersea_texture == NULL && map->undersea_tex_fname != NULL)
  {
    Nprintf("    Loading undersea texture file: \"%s\"\n", map->undersea_tex_fname);
    map->undersea_texture = load_natow_texture(map->undersea_tex_fname);
    if(map->undersea_texture == NULL)
      Neprintf("      Error loading undersea texture file: \"%s\"! (skipped)\n",
	       map->undersea_tex_fname);
  }

  /* If some textures were specified, but weren't loaded, return FALSE */
  if((!map->rock_texture && map->rock_tex_fname) ||
     (!map->ice_texture && map->ice_tex_fname) ||
     (!map->tree_texture && map->tree_tex_fname) ||
     (!map->grass_texture && map->grass_tex_fname) ||
     (!map->sand_texture && map->sand_tex_fname) ||
     (!map->undersea_texture && map->undersea_tex_fname))
    return FALSE;
  else
    return TRUE;
}

/* Generate a map texture, based on smaller textures, and the shape of the map,
   return FALSE if unable to generate texture */
int generate_map_texture(Map *map)
{
  /* The mini-textures */
  NatowTexture *rock_texture, *ice_texture, *tree_texture;
  NatowTexture *grass_texture, *sand_texture, *undersea_texture;
  /* The data in the mini-textures */
  GLubyte ***rock_data = NULL, ***ice_data = NULL, ***tree_data = NULL;
  GLubyte ***grass_data = NULL, ***sand_data = NULL, ***undersea_data = NULL;
  /* The position on each mini-texture (for a point in the map texture) */
  int rock_pos[2], ice_pos[2], tree_pos[2], grass_pos[2], sand_pos[2], undersea_pos[2];
  /* The amount of each mini-texture (for a point in the map texture) */
  float slope, slope_2, slope_4, p;
  float rock_percent = 0.0, ice_percent = 0.0, tree_percent = 0.0;
  float grass_percent = 0.0, sand_percent = 0.0, undersea_percent = 0.0;
  int w, h, c, size_exp;

  /* The properties of the current location */
  float map_pos[2], abs_alt, alt, alt_percent, *norm, map_colour[3];

  /* Check that all the mini-texures are loaded */
  if(!load_mini_textures(map))
  {
    Neprintf("  Warning, not all mini-textures loadable\n");
    return FALSE;
  }

  rock_texture = map->rock_texture;
  ice_texture = map->ice_texture;
  tree_texture = map->tree_texture;
  grass_texture = map->grass_texture;
  sand_texture = map->sand_texture;
  undersea_texture = map->undersea_texture;

  /* Shortcuts to the data */
  if(rock_texture)
    rock_data = rock_texture->data;
  if(ice_texture)
    ice_data = ice_texture->data;
  if(tree_texture)
    tree_data = tree_texture->data;
  if(grass_texture)
    grass_data = grass_texture->data;
  if(sand_texture)
    sand_data = sand_texture->data;
  if(undersea_texture)
    undersea_data = undersea_texture->data;

  map->tex_width = (int)(map->width / map->game->world->tex_x_resolution);
  map->tex_height = (int)(map->height / map->game->world->tex_y_resolution);

  /* Align the texture to be 2^n1 x 2^n2 (round down) */
  frexp(map->tex_width, &size_exp);
  map->tex_width = pow(2, size_exp);
  frexp(map->tex_height, &size_exp);
  map->tex_height = pow(2, size_exp);
  Nprintf("    Map texture size %dx%d\n", map->tex_width, map->tex_height);

  /* Clear out an old texture */
  Nfree(map->texture_data);
  map->texture_data = NULL;

  /* Allocate enough space for the texture */
  map->texture_data = malloc(sizeof(GLubyte) * map->tex_width * map->tex_height * 3);

  Nprintf("    Generating texture elements...\r");
  fflush(stdout);

  /* Go through each texel, and calcuate the colour blended from the different
     textures */
  for(w = 0; w < map->tex_width; w++)
  {
    Nprintf("    Generating texture elements... [%d/%d]\r", w+1, map->tex_width);
    /* This makes the loading display smoother, but it slows it down a lot */
    /*fflush(stdout);*/

    /* The position on the mini-textures this is */
    if(rock_texture)
      rock_pos[0]     = w % rock_texture->width;
    if(ice_texture)
      ice_pos[0]      = w % ice_texture->width;
    if(tree_texture)
      tree_pos[0]     = w % tree_texture->width;
    if(grass_texture)
      grass_pos[0]    = w % grass_texture->width;
    if(sand_texture)
      sand_pos[0]     = w % sand_texture->width;
    if(undersea_texture)
      undersea_pos[0] = w % undersea_texture->width;

    for(h = 0; h < map->tex_height; h++)
    {
      /* Current position on the map this translates to */
      map_pos[0] = (float)w / (float)map->tex_width * map->width;
      map_pos[1] = (float)h / (float)map->tex_height * map->height;

      /* The position on the mini-textures this is */
      if(rock_texture)
	rock_pos[1]     = h % rock_texture->height;
      if(ice_texture)
	ice_pos[1]      = h % ice_texture->height;
      if(tree_texture)      
	tree_pos[1]     = h % tree_texture->height;
      if(grass_texture)
	grass_pos[1]    = h % grass_texture->height;
      if(sand_texture)
	sand_pos[1]     = h % sand_texture->height;
      if(undersea_texture)
	undersea_pos[1] = h % undersea_texture->height;

      /* The map altitude (abs_alt),
	 altitude above sea-level (alt),
	 percentage of altitude range (alt_percent),
	 and normal (norm) at this point */
      abs_alt = get_height(map, map_pos[0], map_pos[1]);
      alt = abs_alt - map->game->sea_level;
      alt_percent = abs_alt / map->max_altitude;
      norm = get_smooth_normal(map, map_pos[0], map_pos[1]);

      slope = norm[1];
      slope_2 = slope*slope;
      slope_4 = slope_2*slope_2;

      /* The percentage of each texture is used in the map texture */
      if(alt < 0.0)
      {
	if(alt < -4.0) /* If deep underwater */
	{
	  rock_percent     = 0.0;
	  sand_percent     = 0.0;
	  undersea_percent = 1.0;
	}
	else
	{
	  p = -alt / 4.0;
	  rock_percent     = (1.0 - p) * (1.0 - slope_4);
	  sand_percent     = (1.0 - p) * slope_4;
	  undersea_percent = p;
	}
	ice_percent      = 0.0;
	tree_percent     = 0.0;
	grass_percent    = 0.0;
      }
      else
      {
	rock_percent     = 1.0 - slope_4;
	ice_percent      = alt_percent * slope_4;
	tree_percent     = 0.0;
	grass_percent    = (1.0 - alt_percent) * slope_4;
	if(alt < 2.0) /* If under 2m from the sea */
	{
	  sand_percent   = slope_4;
	  if(alt > 1.0)
	    grass_percent = (alt - 1.0) / 2.0 * (1.0 - alt_percent) * slope_4;
	  else
	    grass_percent = 0.0;
	}
	else
	  sand_percent     = 0.0;
	undersea_percent = 0.0;
      }

      /* Blend the textures together using floating point */
      for(c = 0; c < 3; c++)
      {
	map_colour[c] = 0.0;
	if(rock_data)
	  map_colour[c] += rock_percent *
	    (float)rock_data[rock_pos[0]][rock_pos[1]][c] / 255.0;
	if(ice_data)
	  map_colour[c] += ice_percent *
	    (float)ice_data[ice_pos[0]][ice_pos[1]][c] / 255.0;
	if(tree_data)
	  map_colour[c] += tree_percent *
	    (float)tree_data[tree_pos[0]][tree_pos[1]][c] / 255.0;
	if(grass_data)
	  map_colour[c] += grass_percent *
	    (float)grass_data[grass_pos[0]][grass_pos[1]][c] / 255.0;
	if(sand_data)
	  map_colour[c] += sand_percent *
	    (float)sand_data[sand_pos[0]][sand_pos[1]][c] / 255.0;
	if(undersea_data)
	  map_colour[c] += undersea_percent *
	    (float)undersea_data[undersea_pos[0]][undersea_pos[1]][c] / 255.0;
      }

      /* Clip the colours to valid ranges */
      for(c = 0; c < 3; c++)
	if(map_colour[c] > 1.0)
	  map_colour[c] = 1.0;

      /* Convert this value to an unsigned byte */
      for(c = 0; c < 3; c++)
	map->texture_data[h*map->tex_width*3 + 3*w + c] = (GLubyte)(map_colour[c] * 255.0);
    }
  }

  Nprintf("    Generating texture elements... [done]               \n");
  fflush(stdout);

  /* Delete mini textures to save space */
  /* NOTE: Use these to regen the main texture later?
     Probably too expensive to do */
  free_natow_texture(map->rock_texture);
  free_natow_texture(map->ice_texture);
  free_natow_texture(map->tree_texture);
  free_natow_texture(map->grass_texture);
  free_natow_texture(map->sand_texture);
  free_natow_texture(map->undersea_texture);
  map->rock_texture = NULL;
  map->ice_texture = NULL;
  map->tree_texture = NULL;
  map->grass_texture = NULL;
  map->sand_texture = NULL;
  map->undersea_texture = NULL;

  regen_map_texture(map);

  /* Remove the local copy of the texture */
  /* NOTE: This means the main texture can't be easily regenerated */
  /* I keep this texture in order to add 'dark traces' on the ground
       after explosions (see calc_map_trace() in explosion.c) (YP) */
  /*
  Nfree(map->texture_data);
  map->texture_data = NULL;
  */

  return TRUE;
}

/* Remake the map texture, using the values present */
int regen_map_texture(Map *map)
{
  /* Remove the old texture */
  glDeleteTextures(1, &map->texture);

  /* Generate the texture */
  glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
  glGenTextures(1, &map->texture);
  glBindTexture(GL_TEXTURE_2D, map->texture);

  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, map->game->world->texture_mode);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, map->game->world->texture_mode);

  glTexImage2D(GL_TEXTURE_2D,
	       0,
	       3,
	       map->tex_width,
	       map->tex_height,
	       0,
	       GL_RGB,
	       GL_UNSIGNED_BYTE,
	       map->texture_data);

  /* Make this texture cover the entire map */
  map->tex_scale_x = 1.0 / (map->size_x - 1.0);
  map->tex_scale_y = 1.0 / (map->size_y - 1.0);

  return TRUE;
}
