/* 
 * NATOW - Tank to tank warfare in a 3D fully destructable world
 *
 * Copyright (C) 2000-2001
 *                     Robert Ancell  <bob27@users.sourceforge.net>
 *                     Yannick Perret <hexasoft@users.sourceforge.net>
 *                     Joel Pitt      <extropy@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "pathfind.h"
#include "world.h"
#include "game.h"
#include "text.h"
#include "tools.h"

static pathnode *allocate_node(void);
static pathnode *split_path(Game *game, pathnode* path, int fineness);
static int trace(Map *map, pathnode* path, pathnode* end);
static void try_dry_pos(Map *map, int x, int y);
static void draw_node(World *world, pathnode *node);

static float pathfind_inc = 1.0; /* smaller increment gives closer path, takes longer */
static float path_find_fineness = 2.0; /* lower is finer */

/* Allocate a path node */
static pathnode *allocate_node(void)
{
  pathnode *node;

  node = malloc(sizeof(pathnode));
  if(!node)
  {
    Neprintf("Error: Unable to allocate space for path node\n");
    return NULL;
  }

  node->x = 0.0;
  node->y = 0.0;
  node->next = NULL;
  node->last = NULL;

  return node;
}

/* make a path from player to (x,y)
   returns a linked list of path nodes if a path can be found,
   otherwise returns NULL */
pathnode *find_path(Player *player, float x, float y)
{
  int fineness;
  float dx, dy, px, py;

  int i;

  pathnode *path, *node;
  pathnode *current_node;

  /* The postion of the player */
  px = player->pos[0];
  py = player->pos[2];

  dx = x - px;
  dy = y - py;
  path = allocate_node();
  node = path;
  for(i = 0; i < 10; i++)
  {
    node->x = px + dx * (float)i / 10.0;
    node->y = py + dy * (float)i / 10.0;

    node->next = allocate_node();
    node = node->next;
  }
  node->x = x;
  node->y = y;

  return path;

  path = allocate_node();
  path->x = player->pos[0];
  path->y = player->pos[2];
  
  path->next = allocate_node();
  current_node = path->next;
  
  current_node->last = path;
  current_node->x = x;
  current_node->y = y;
  
  dx = path->x - x;
  dy = path->y - y;
  fineness = (int) (sqrt((dx*dx) + (dy*dy)) / 2);
  
  /*  if(!split_path(player->world, path, fineness))
  {
    free_path(path);
    return NULL;
	} */

  return path;
}

/* Used to find a dry position on the map */
static int have_dry_pos, dry_x, dry_y;

/* Make a path to dry ground */
pathnode *find_dry_path(Player *player)
{
  Game *game;
  Map *map;
  pathnode *path, *node;
  int px, py, map_width, map_height;
  int dr, dr2, max_dim;

  if(!player || !player->game || !player->game->map)
    return NULL;

  game = player->game;
  map = game->map;

  /* Make a node right where they are */
  path = allocate_node();
  path->x = player->pos[0];
  path->y = player->pos[2];

  /* Make a node on the nearest dry ground */
  /* The nearest map co-ordinate of the player */
  px = (int)(path->x / map->x_resolution + 0.5);
  py = (int)(path->y / map->y_resolution + 0.5);
  map_width = map->size_x;
  map_height = map->size_y;
  max_dim = map_width > map_height ? map_width : map_height;

  have_dry_pos = FALSE;
  dry_x = px;
  dry_y = py;

  for(dr = 1; dr < max_dim; dr++)
    for(dr2 = 0; dr2 < dr; dr2++)
    {
      try_dry_pos(map, px + dr, py + dr2);
      try_dry_pos(map, px + dr, py - dr2);
      try_dry_pos(map, px - dr, py + dr2);
      try_dry_pos(map, px - dr, py - dr2);

      try_dry_pos(map, px + dr2, py + dr);
      try_dry_pos(map, px - dr2, py + dr);
      try_dry_pos(map, px + dr2, py - dr);
      try_dry_pos(map, px - dr2, py - dr);

      if(have_dry_pos)
      {
	dr2 = dr;
	dr = max_dim;
      }
    }

  node = path->next = allocate_node();
  node->x = dry_x;
  node->y = dry_y;
  node->last = path;

  /* NOTE: Should now split path so is not too steep */

  return path;
}

/* See if this position is dry (and in map bounds) */
static void try_dry_pos(Map *map, int x, int y)
{
  if(x < 0 || y < 0 || x > map->size_x || y > map->size_y)
    return;

  if(map->pos[x][y][1] > map->game->sea_level)
  {
    have_dry_pos = TRUE;
    dry_x = x;
    dry_y = y;
  }
}

/* Returns true if there is a straight path from start to end
   without any water in the way */
static int trace(Map *map, pathnode* start, pathnode* end) {

  float u[2], length;
  float temp_pos[2];
  float *lowx, *lowy, *hix, *hiy;
  
  /* Calc. unit vector from begin to end */
  u[0] = start->x - end->x;
  u[1] = start->y - end->y;
  length = sqrt( (u[0] * u[0]) + (u[1] * u[1]) );
  u[0] /= length;
  u[1] /= length;

  temp_pos[0] = start->x;
  temp_pos[1] = start->y;

  if (start->x < end->x) {
	lowx = &(temp_pos[0]);
	hix = &(end->x);
  } else {
	lowx = &(end->x);
	hix = &(temp_pos[0]);
  }
  if (start->y < end->y) {
	lowy = &(temp_pos[1]);
	hiy = &(end->y);
  } else {
	lowy = &(end->y);
	hiy = &(temp_pos[1]);
  }

  while (*lowx < *hix && *lowy < *hiy) {

	temp_pos[0] += u[0];
	temp_pos[1] += u[1];
  
	try_dry_pos(map, (int) temp_pos[0], (int) temp_pos[1]);
	if (have_dry_pos != TRUE) return FALSE;

  }

  return TRUE;

}

static pathnode *split_path(Game *game, pathnode* path, int fineness)
{
  pathnode* new_node, *path_end;
  float dx, dy;

  if(!fineness || !path->next)
    return NULL;
  path_end = path->next;
  
  new_node = malloc(sizeof(pathnode));
  new_node->next = path_end;
  new_node->last = path;
    
  if (path->x >= path_end->x) {
    dx = path->x - path_end->x;
	new_node->x =  path_end->x + dx / 2;
  }
  else {
	dx = (path_end->x - path->x);
	new_node->x =  path->x + dx / 2;
  }
  
  if (path->y >= path_end->y) {
	dy = (path->y - path_end->y);
	new_node->y =  path_end->y + dy / 2;
  }
  else {
    dy = (path_end->y - path->y);
	new_node->y =  path->y + dy / 2;
  }

  path->next = new_node;
  path_end->last = new_node;

  /* If the new node is in water we've got to move it */
  if (game->sea_level > get_height(game->map, new_node->x, new_node->y)) {
	/* Try moving point perpendicularly both ways, first not in water wins. */

	float dx2, dy2, c, c2; /* dx and dy are used for calculating the gradient
							  and c is the intercept */
	float tempx, tempy; /* one location is kept in new_node and the other in
						   tempx and tempy */
	int option = 0;

	dx2 = dx; dy2 = dy;
	tempx = new_node->x; tempy = new_node->y;

	/* find gradients of the two perpendicular vectors and the intercepts
	   for the equation y = mx + c */
	dx = dx * -1;
	dy2 = dy2 * -1;
	c = new_node->y - (dy/dx);
	c2 = new_node->y - (dy2/dx2);

  	while (option == 0) {

	  if (new_node->x < game->map->width && new_node->y < game->map->height) {
		if (game->sea_level > get_height(game->map, new_node->x, new_node->y)) {
		  new_node->x = new_node->x + pathfind_inc;
		  new_node->y = (dy/dx) * new_node->x + c;
		} else option = 1;
	  }
	 
	  if (tempx < game->map->width && tempy < game->map->height) {
		if (game->sea_level > get_height(game->map, tempx, tempy)) {
		  tempx = tempx + pathfind_inc;
		  tempy = (dy2/dx2) * tempx + c2;
		} else option = 2;
	  } else if (new_node->x < game->map->width && new_node->y < game->map->height) option = -1;
	}

	if (option == 2) {
	  new_node->x = tempx;
	  new_node->y = tempy;
	} else if (option == -1) return NULL;
  }

  fineness--;
  if (split_path(game, path, fineness) == NULL ||
	  split_path(game, new_node, fineness) == NULL) return NULL;
  
  return path;
  
}

/* Remove the first node */
pathnode *pop_node(pathnode *path)
{
  pathnode *node1;

  node1 = path->next;
  if(node1)
    node1->last = NULL;

  Nfree(path);

  return node1;
}

/* Draw a path onto the map */
void draw_path(World *world, pathnode *path)
{
  pathnode *node;

  glPushAttrib(GL_LIGHTING_BIT | GL_DEPTH_BUFFER_BIT);

  glDisable(GL_LIGHTING);
  glDisable(GL_DEPTH_TEST);

  node = path;
  while(node)
  {
    draw_node(world, node);
    node = node->next;
  }

  glPopAttrib();
}

/* Draw a node and the link to the next node (if there is one) */
static void draw_node(World *world, pathnode *node)
#define NODE_HEIGHT 5.0
{
  float h, dx, dy, dz, d, *cpos;
  pathnode *n2;

  if(!node || !world->game || !world->game->round)
    return;

  /* NOTE: Should leave now if node not visible,
     need camera->dir */

  /* Get the height of this node + a constant so is above map */
  h = get_height(world->game->map, node->x, node->y) + NODE_HEIGHT;

  /* The distance to this point */
  cpos = world->game->round->camera->pos;
  dx = node->x - cpos[0];
  dy = h - cpos[1];
  dz = node->y - cpos[2];
  d = sqrt(dx*dx + dy*dy + dz*dz);

  glEnable(GL_POINT_SMOOTH);
  glPointSize(1000.0 / d);
  glLineWidth(100.0 / d);

  /* Draw a blip at the node */
  glBegin(GL_POINTS);
  glColor3f(1.0, 0.0, 0.0);
  glVertex3f(node->x, h, node->y);
  glEnd();

  /* Draw a connecting line between this node and the next */
  glColor3f(1.0, 1.0, 0.0);
  if((n2 = node->next))
  {
    glBegin(GL_LINES);

    glVertex3f(node->x, h, node->y);
    glVertex3f(n2->x, get_height(world->game->map, n2->x, n2->y) + NODE_HEIGHT, n2->y);

    glEnd();
  }
}

int free_path(pathnode* path) {
  
  pathnode *tempnode;

  while (path) {
	tempnode = path->next;
	free(path);
	path = tempnode;
  }

  return 0;
}
