/* 
 * NATOW - Tank to tank warfare in a 3D fully destructable world
 *
 * Copyright (C) 2000-2001
 *                     Robert Ancell  <bob27@users.sourceforge.net>
 *                     Yannick Perret <hexasoft@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* turn.c
   Manage turns in each round */

#include "turn.h"
#include "text.h"
#include "game.h"
#include "round.h"
#include "player.h"
#include "player_list.h"
#include "camera.h"

/* Start a new turn */
int start_turn(Round *round)
{
  PlayerEntry *entry;

  if(!round)
    return FALSE;

  /* Move onto the next turn */
  round->turn++;
  if(round->turn > round->max_turns)
    return FALSE;

  Nprintf("Starting turn %d...\n", round->turn);

  /* Randomize the players */
  randomize_player_list(round->players);

  /* The player who starts this round
     i.e. Find the first alive player */
  entry = round->players->first;
  while(entry && entry->player->player_state != PLAYER_ALIVE)
    entry = entry->next;

  /* No alive players -- the round will end soon */
  if(!entry)
  {
    round->turn = round->max_turns + 1;
    return FALSE;
  }

  /* The player to start this turn */
  round->current_player_entry = entry;
  round->current_player = entry->player;

  /* Have the camera target this new player */
  round->camera->target_pos = round->current_player->pos;

  return TRUE;
}

/* Gets called when a player is done */
int player_end_turn(Player *player)
{
  Game *game;
  Round *round;
  PlayerEntry *pentry;
  Player *player_ptr;

  if(!player || !player->game || !player->game->round)
    return FALSE;

  game = player->game;
  round = game->round;

  /* If in single player mode, finished ==> fire */
  if(round->game->mode == SINGLE_PLAYER)
  {
    if(player->player_state == PLAYER_ALIVE)
      fire_player(player);
    return TRUE;
  }

  /* Put the brakes on */
  if(player->engine_on)
  {
    set_player_brakes(player, TRUE);
    stop_player_engine(player);
    player->has_driven = TRUE;
  }

  /* Cycle to the next living player */
  if(round->current_player_entry)
    do
    {
      round->current_player_entry = round->current_player_entry->next;
    } while(round->current_player_entry &&
	    round->current_player_entry->player->player_state != PLAYER_ALIVE);

  /* If haven't run out of players */
  if(round->current_player_entry)
  {
    /* The new current player */
    round->current_player = round->current_player_entry->player;

    /* Have the camera target this new player */
    round->camera->target_pos = round->current_player->pos;
  }
  else
  {
    round->current_player = NULL;

    /* Fire all the players */
    pentry = round->players->first;
    while(pentry)
    {
      player_ptr = pentry->player;

      /* If alive, and didn't move last turn */
      if((player_ptr->player_state == PLAYER_ALIVE) && !player_ptr->has_driven)
	fire_player(player_ptr);

      /* They have more chances to drive */
      player_ptr->has_driven = FALSE;

      pentry = pentry->next;
    }

    /* TODO: Zoom out to see the whole map */
  }

  return TRUE;
}

/* See if the current turn is finished, return TRUE if finished */
int test_turn_end(Round *round)
{
  /* If turn is zero, first turn hasn't started */
  if(round->turn == 0)
    return FALSE;

  /* If not everyone has finished their turn */
  if(round->current_player_entry)
    return FALSE;

  /* If any projectiles or explosions are still to finish */
  if(round->projectile_list || round->blast_list)
    return FALSE;

  Nprintf("Ending turn %d...\n", round->turn);

  /* Start the next turn */
  start_turn(round);

  /* It appears this turn is finished */
  return TRUE;
}
