/* 
 * NATOW - Tank to tank warfare in a 3D fully destructable world
 *
 * Copyright (C) 2000-2001
 *                     Robert Ancell  <bob27@users.sourceforge.net>
 *                     Yannick Perret <hexasoft@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* world.c
   The main structure, everything is based from here */

#include <stdio.h>
#include <malloc.h>

#include "world.h"
#include "game.h"

#include "vector.h"
#include "map.h"
#include "camera.h"
#include "text.h"
#include "draw.h"
#include "2d.h"
#include "shade.h"
#include "about.h"
#include "plasma.h"
#include "start.h"
#include "tools.h"
#include "keymap.h"
#include "display_stack.h"
#include "tank.h"
#include "munition.h"
#include "satellite.h"

/* The default display/update etc functions */
#include "draw.h"
#include "input.h"
#include "update.h"
#include "cursor.h"

/* Make the base world */
World *new_world(void)
{
  World *world = malloc(sizeof(World));

  if(!world)
  {
    Neprintf("    Unable to allocate space for world");
    return NULL;
  }

  world->game = NULL;

  world->total_elapsed_time = 0.0;

  world->use_tutorial = FALSE;

  /* Set some valid defaults */
  world->display_func = display;
  world->reshape_func = reshape;
  world->keyboard_func = keyboard;
  world->motion_func = motion;
  world->update_func = update_world;
  world->move_func = manage_mouse_cursor;
  world->mouse_func = mouse_event;
  world->entry_func = enter_leave;

  world->levels = NULL;

  world->LOD_boundary1 = 10;
  world->LOD_boundary2 = 20;
  world->LOD_boundary3 = 40;
  /* The accuracy of the texture */
  /* Accurate to 0.5m */
  world->tex_x_resolution = 0.5;
  world->tex_y_resolution = 0.5;

  world->cursor_inside      = TRUE;
  world->paused_from_cursor = FALSE;

  world->display_stack = NULL;
  world->shade_mode    = NATOW_SHADE_NONE;
  world->shade_level   = 0.;
  /* Start on the main menu */
  set_display_mode(world, NATOW_DISPLAY_MAIN_MENU);

  /* The list of tanks */
  world->ntanks = 0;
  world->tank_list = NULL;

  /* The list of munitions */
  world->nmunitions = 0;
  world->munition_list = NULL;

  world->player_list = NULL;

  world->low_detail = FALSE;

  world->is_smooth = TRUE;
  world->is_fog = FALSE;
  world->is_pin = FALSE;
  world->is_grid = FALSE;
  world->is_light = TRUE;
  world->is_water = FALSE;
  world->is_texture = TRUE;
  world->is_sky = TRUE;
  world->is_bird = TRUE;
  world->is_shading = TRUE;

  world->sun   = NULL;
  world->moon1 = NULL;
  world->moon2 = NULL;

  world->smoke_percent = 100;
  world->plasma_detail = 2;
  world->texture_mode = GL_LINEAR;

  /* !!note: temp */
  world->fps = 0.0;
  world->max_dt = 1000.0;

  world->video_size_x = 640;
  world->video_size_y = 480;
  world->size_x = 640;
  world->size_y = 480;
  world->fullscreen = FALSE;

  world->lists.sky = 0;
  world->lists.skirt = 0;
  world->lists.lakes = 0;

  world->key_list = get_keymap();

  return world;
}

/* Free everything */
int free_world(World *world)
{
  if(!world)
    return FALSE;

  free_game(world->game);
  clear_level_list(world);
  clear_display_stack(world);
  free_tanks(world);
  clear_munition_list(world);
  remove_players(world);
  free_satellite(world->sun);
  free_satellite(world->moon1);
  free_satellite(world->moon2);
  free_2d(world);
  free_plasma(world);
  free_start(world);
  free_shading(world);
  free_about(world);
  free_cursor(world);


  return TRUE;
}
